use crate::memory::io::BusIO;
use bitfield_struct::bitfield;

#[derive(Debug, Default)]
pub struct APU {
    regs: [u8; 0x20],
    wave: [u8; 0x10],
    amc: AMC,
    pan: SndPan,
    vol: Vol,
}

impl BusIO for APU {
    fn read(&self, addr: usize) -> Option<u8> {
        match addr {
            0xff10..=0xff23 => self.regs.read(addr - 0xff10),
            // 0xff10 => todo!("Channel 1 sweep"),
            // 0xff11 => todo!("Channel 1 length timer & duty cycle"),
            // 0xff12 => todo!("Channel 1 volume & envelope"),
            // 0xff13 => todo!("Channel 1 period low [write-only]"),
            // 0xff14 => todo!("Channel 1 period high & control [trigger, length enable, period]"),
            0xff24 => Some(self.vol.into_bits()),
            0xff25 => Some(self.pan.into_bits()),
            0xff26 => Some(self.amc.into_bits()),
            0xff30..=0xff3f => self.wave.read(addr - 0xff30),
            0xff76 => todo!("PCM12"),
            0xff77 => todo!("PCM34"),
            _ => None,
        }
    }

    fn write(&mut self, addr: usize, data: u8) -> Option<()> {
        #[allow(clippy::unit_arg)]
        match addr {
            0xff10..0xff14 => todo!("Channel 1"),
            0xff16..=0xff19 => todo!("Channel 2"),
            0xff1a..=0xff1e => todo!("Channel 3"),
            0xff20..=0xff23 => todo!("Channel 4"),
            0xff24 => Some(self.vol = Vol::from_bits(data)),
            // Sound panning
            0xff25 => Some(self.pan = SndPan::from_bits(data)),
            // Audio Master Control
            0xff26 => Some(self.amc.set_power(data & 0x80 != 0)),
            // Wave table memory
            0xff30..=0xff3f => self.wave.write(addr - 0xff30, data),
            0xff76 => todo!("PCM12"),
            0xff77 => todo!("PCM34"),
            _ => None,
        }
    }
}

#[derive(Debug, Default)]
pub struct Channel {
    pub period: u16,
    pub duty_step: u8,
}

#[bitfield(u8)]
/// Audio Master Control register
pub struct AMC {
    /// Set when CH1 is playing. Read-only.
    ch1_on: bool,
    /// Set when CH2 is playing. Read-only.
    ch2_on: bool,
    /// Set when CH3 is playing. Read-only.
    ch3_on: bool,
    /// Set when CH4 is playing. Read-only.
    ch4_on: bool,
    #[bits(3)] // 3 bits are unused
    _reserved: (),
    /// If false, all channels are muted. Read-writable.
    power: bool,
}

#[bitfield(u8)]
/// Sound Panning register. Setting `cNS` to true will send channel `N` to speaker `S`
pub struct SndPan {
    c1r: bool,
    c2r: bool,
    c3r: bool,
    c4r: bool,
    c1l: bool,
    c2l: bool,
    c3l: bool,
    c4l: bool,
}

#[bitfield(u8)]
/// Volume control and VIN panning register
pub struct Vol {
    #[bits(3)]
    /// The volume of the right output
    right: u8,
    /// Whether VIN is sent to the right output
    vin_right: bool,
    #[bits(3)]
    /// The volume of the left output
    left: u8,
    /// Whether VIN is sent to the left output
    vin_left: bool,
}

pub mod channels {
    // TODO: remove
    #![allow(dead_code)]
    //! Channel controllers. Each channel controller has a function that takes a bus and outputs

    static WAVES: [[u8; 8]; 4] = [
        [0xf, 0xf, 0xf, 0xf, 0xf, 0xf, 0x0, 0xf],
        [0xf, 0xf, 0xf, 0xf, 0xf, 0xf, 0x0, 0x0],
        [0xf, 0xf, 0xf, 0xf, 0x0, 0x0, 0x0, 0x0],
        [0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0xf, 0xf],
    ];

    #[derive(Debug, Default)]
    pub struct Pulse {
        /// Ticks up from the initial value, and triggers a duty step when it reaches 0x800
        period: u16,
        duty_step: u8,
        duty_cycle: u8,
        output: u8,
    }

    impl Pulse {
        pub fn tick(&mut self) {
            self.period += 1;
            if self.period > 0x7ff {
                // move the duty step
                self.duty_step = (self.duty_step + 1) % 8;
            }
        }
    }
}
