//! Disassembler for SM83 machine code

#[derive(Clone, Copy, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Insn {
    #[default]
    /// No operation
    Nop,
    Stop,
    Halt,
    /// [Register](R8) to [register](R8) transfer
    Ld(R8, R8),
    /// Load an [8-bit](R8) immediate: `ld r8, #u8`
    LdImm(R8),
    /// Load a [16-bit](R16) immediate: `ld r16, #u16`
    LdImm16(R16),
    /// Load from [R16Indirect] into register [`a`](R8::A)
    LdInd(R16Indirect),
    /// Store register [`a`](R8::A) into [R16Indirect]
    StInd(R16Indirect),
    /// Load [`a`](R8::A) from 0xff00+[`c`](R8::C): `ld a, [c]`
    LdC,
    /// Store [`a`](R8::A) into 0xff00+[`c`](R8::C): `ld [c], a`
    StC,
    /// Load [`a`](R8::A) from absolute address: `ld a, [addr]`
    LdAbs,
    /// Store [`a`](R8::A) to absolute address: `ld [addr], a`
    StAbs,
    /// Store [stack pointer](R16::SP) to absolute address: `ld [addr], sp`
    StSpAbs,
    /// Load [`HL`](R16::HL) into SP
    LdSpHl,
    /// Load [`HL`](R16::HL) with SP + [u8]
    LdHlSpRel,
    /// Load from himem at immediate offset: `ldh a, [0xff00 + #u8]`
    Ldh,
    /// Store to himem at immediate offset: `ldh [0xff00 + #u8], a`
    Sth,

    /// Increment the value in the [register](R8)
    Inc(R8),
    /// Increment the value in the [register](R16)
    Inc16(R16),
    /// Decrement the value in the [register](R8)
    Dec(R8),
    /// Decrement the value in the [register](R16)
    Dec16(R16),

    /// Rotate left [`a`](R8::A)
    Rlnca,
    /// Rotate right [`a`](R8::A)
    Rrnca,
    /// Rotate left [`a`](R8::A) [through carry flag](https://rgbds.gbdev.io/docs/v0.7.0/gbz80.7#RLA)
    Rla,
    /// Rotate Right [`a`](R8::A) [through carry flag](https://rgbds.gbdev.io/docs/v0.7.0/gbz80.7#RRA)
    Rra,
    /// daa ; Decimal-adjust accumulator using sub and half-carry flag
    Daa,
    /// cpl ; Complement [`a`](R8::A)
    Cpl,
    /// scf ; Set carry flag
    Scf,
    // ccf ; Complement carry flag
    Ccf,

    /// add [`a`](R8::A), [R8]
    Add(R8),
    /// add[`HL`](R16::HL), [R16]
    Add16HL(R16),
    /// add #[u8], [`a`](R8::A)
    AddI,
    /// add [`sp`](R16::SP), #[i8]
    Add16SpI,
    /// adc [R8], [`a`](R8::A)
    Adc(R8),
    /// adc #[u8], [`a`](R8::A)
    AdcI,
    /// sub [R8], [`a`](R8::A)
    Sub(R8),
    /// sub #[u8], [`a`](R8::A)
    SubI,
    /// sbc [R8], [`a`](R8::A)
    Sbc(R8),
    /// sbc #[u8], [`a`](R8::A)
    SbcI,
    /// and [R8], [`a`](R8::A)
    And(R8),
    /// and #[u8], [`a`](R8::A)
    AndI,
    /// xor [R8], [`a`](R8::A)
    Xor(R8),
    /// xor #[u8], [`a`](R8::A)
    XorI,
    /// or  [R8], [`a`](R8::A)
    Or(R8),
    /// or  #[u8], [`a`](R8::A)
    OrI,
    /// cp  [R8], [`a`](R8::A)
    Cp(R8),
    /// cp  #[u8], [`a`](R8::A)
    CpI,

    /// push [r16](R16Stack)
    Push(R16Stack),
    /// pop [r16](R16Stack)
    Pop(R16Stack),

    /// jr #[i8]
    Jr,
    /// jr [Cond], #[i8]
    Jrc(Cond),
    /// jp #[u16]
    Jp,
    /// jp [Cond], #[u16]
    Jpc(Cond),
    /// jp [HL](R16::HL)
    JpHL,
    /// call #[u16]
    Call,
    /// call [Cond], #[u16]
    Callc(Cond),
    /// ret
    Ret,
    /// ret [Cond] ;
    Retc(Cond),

    /// di ; Disable Interrupts
    Di,
    /// ei ; Enable Interrupts
    Ei,
    /// reti ; Return from interrupt, re-enabling interrupts
    Reti,
    /// rst #vec ; Call one of the reset vectors in the first page
    Rst(u8),
    /// (cb) NN ; Next instruction is a [Prefixed] instruction
    PrefixCB,
    Invalid(u8),
}
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Prefixed {
    Rlc(R8),
    Rrc(R8),
    Rl(R8),
    Rr(R8),
    Sla(R8),
    Sra(R8),
    Swap(R8),
    Srl(R8),
    Bit(R8, U3),
    Res(R8, U3),
    Set(R8, U3),
}
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum U3 {
    _0,
    _1,
    _2,
    _3,
    _4,
    _5,
    _6,
    _7,
}
impl From<u8> for Insn {
    fn from(value: u8) -> Self {
        // low, middle, and high nibbles of the instruction
        let (low, mid, high) = (value.low(), value.mid(), value.high());
        // R16 specifier bits, and alternate-instruction bit
        let (r16, alt) = (mid >> 1, mid & 1 == 0);
        match high {
            0 => match (low, mid) {
                (0, 0) => Insn::Nop,
                (0, 1) => Insn::LdSpHl,
                (0, 2) => Insn::Stop,
                (0, 3) => Insn::Jr,
                (0, _) => Insn::Jrc((mid & 3).into()),
                (1, _) if alt => Insn::LdImm16(r16.into()),
                (1, _) => Insn::Add16HL(r16.into()),
                (2, _) if alt => Insn::StInd(r16.into()),
                (2, _) => Insn::LdInd(r16.into()),
                (3, _) if alt => Insn::Inc16(r16.into()),
                (3, _) => Insn::Dec16(r16.into()),
                (4, dst) => Insn::Inc(dst.into()),
                (5, dst) => Insn::Dec(dst.into()),
                (6, dst) => Insn::LdImm(dst.into()),
                (7, 0) => Insn::Rlnca,
                (7, 1) => Insn::Rrnca,
                (7, 2) => Insn::Rla,
                (7, 3) => Insn::Rra,
                (7, 4) => Insn::Daa,
                (7, 5) => Insn::Cpl,
                (7, 6) => Insn::Scf,
                (7, 7) => Insn::Ccf,
                _ => unreachable!(),
            },
            1 if value == 0b01110110 => Insn::Halt,
            1 => Insn::Ld(mid.into(), low.into()),
            2 => match mid {
                0 => Insn::Add(low.into()),
                1 => Insn::Adc(low.into()),
                2 => Insn::Sub(low.into()),
                3 => Insn::Sbc(low.into()),
                4 => Insn::And(low.into()),
                5 => Insn::Xor(low.into()),
                6 => Insn::Or(low.into()),
                7 => Insn::Cp(low.into()),
                _ => unreachable!(),
            },
            3 => match (low, mid) {
                (0, 0..=3) => Insn::Retc((mid & 3).into()),
                (0, 4) => Insn::Sth,
                (0, 5) => Insn::Add16SpI,
                (0, 6) => Insn::Ldh,
                (0, 7) => Insn::LdHlSpRel,
                (1, _) if alt => Insn::Pop(r16.into()),
                (1, 1) => Insn::Ret,
                (1, 3) => Insn::Reti,
                (1, 5) => Insn::JpHL,
                (1, 7) => Insn::LdSpHl,
                (2, 0..=3) => Insn::Jpc((mid & 3).into()),
                (2, 4) => Insn::StC,
                (2, 5) => Insn::StAbs,
                (2, 6) => Insn::LdC,
                (2, 7) => Insn::LdAbs,
                (3, 0) => Insn::Jp,
                (3, 1) => Insn::PrefixCB,
                (3, 6) => Insn::Di,
                (3, 7) => Insn::Ei,
                (3, _) => Insn::Invalid(value),
                (4, 0..=3) => Insn::Callc((mid & 3).into()),
                (4, _) => Insn::Invalid(value),
                (5, 1) => Insn::Call,
                (5, _) if alt => Insn::Push(r16.into()),
                (5, _) => Insn::Invalid(value),
                (6, 0) => Insn::AddI,
                (6, 1) => Insn::AdcI,
                (6, 2) => Insn::SubI,
                (6, 3) => Insn::SbcI,
                (6, 4) => Insn::AndI,
                (6, 5) => Insn::XorI,
                (6, 6) => Insn::OrI,
                (6, 7) => Insn::CpI,
                (7, rst) => Insn::Rst(rst),
                _ => unreachable!(),
            },
            _ => unreachable!(),
        }
    }
}
impl From<u8> for Prefixed {
    fn from(value: u8) -> Self {
        let r8 = value.low().into();
        match (value.high(), value.mid()) {
            (0, 0) => Self::Rlc(r8),
            (0, 1) => Self::Rrc(r8),
            (0, 2) => Self::Rl(r8),
            (0, 3) => Self::Rr(r8),
            (0, 4) => Self::Sla(r8),
            (0, 5) => Self::Sra(r8),
            (0, 6) => Self::Swap(r8),
            (0, 7) => Self::Srl(r8),
            (1, b3) => Self::Bit(r8, b3.into()),
            (2, b3) => Self::Res(r8, b3.into()),
            (3, b3) => Self::Set(r8, b3.into()),
            _ => unreachable!(),
        }
    }
}
impl From<u8> for U3 {
    fn from(value: u8) -> Self {
        match value & 7 {
            0 => Self::_0,
            1 => Self::_1,
            2 => Self::_2,
            3 => Self::_3,
            4 => Self::_4,
            5 => Self::_5,
            6 => Self::_6,
            7 => Self::_7,
            _ => unreachable!(),
        }
    }
}
impl From<U3> for u8 {
    fn from(value: U3) -> Self {
        value as _
    }
}

/// Note: Meant only for implementation on numeric types
pub trait U8ins: Sized {
    /// Gets the Nth octal digit of Self
    fn oct<const N: u8>(self) -> u8;
    /// Gets the low octal digit of self
    /// ```binary
    /// 00000111 -> 111
    /// ```
    fn low(self) -> u8 {
        self.oct::<0>()
    }
    /// Gets the middle octal digit of self, also known as B3/TGT3
    /// ```binary
    /// 00111000 -> 111
    /// ```
    fn mid(self) -> u8 {
        self.oct::<1>()
    }
    /// Gets the high octal digit of self
    /// ```binary
    /// 11000000 -> 011
    /// ```
    fn high(self) -> u8 {
        self.oct::<2>()
    }
}
impl U8ins for u8 {
    fn oct<const N: u8>(self) -> u8 {
        self >> (N * 3) & 0b111
    }
}

/// 8 bit register or RAM access
/// 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7
/// --|---|---|---|---|---|---|--
/// B | C | D | E | H | L |*HL| A
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum R8 {
    B,
    C,
    D,
    E,
    H,
    L,
    HLmem,
    A,
}
/// 16 bit register
///  0  | 1  | 2  | 3
/// ----|----|----|----
///  BC | DE | HL | SP
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum R16 {
    BC,
    DE,
    HL,
    SP,
}
/// 16 bit operand to push and pop
///  0  | 1  | 2  | 3
/// ----|----|----|----
///  BC | DE | HL | AF
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum R16Stack {
    BC,
    DE,
    HL,
    AF,
}
/// 16 bit memory access through pointer
///  0  | 1  | 2  | 3
/// ----|----|----|----
///  BC | DE | HL+| HL-
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum R16Indirect {
    /// Address in [R16::BC]
    BC,
    /// Address in [R16::DE]
    DE,
    /// Address in [R16::HL]. Post-increment HL.
    HLI,
    /// Address in [R16::HL]. Post-decrement HL.
    HLD,
}

///  0  | 1  | 2  | 3
/// ----|----|----|----
///  NZ | Z  | NC | C
#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum Cond {
    NZ,
    Z,
    NC,
    C,
}

mod conv {
    use super::*;
    impl From<u8> for R8 {
        fn from(value: u8) -> Self {
            match value {
                0 => Self::B,
                1 => Self::C,
                2 => Self::D,
                3 => Self::E,
                4 => Self::H,
                5 => Self::L,
                6 => Self::HLmem,
                7 => Self::A,
                n => panic!("Index out of bounds: {n} > 7"),
            }
        }
    }
    impl From<u8> for R16 {
        fn from(value: u8) -> Self {
            match value {
                0 => Self::BC,
                1 => Self::DE,
                2 => Self::HL,
                3 => Self::SP,
                n => panic!("Index out of bounds: {n} > 3"),
            }
        }
    }
    impl From<u8> for R16Stack {
        fn from(value: u8) -> Self {
            match value {
                0 => Self::BC,
                1 => Self::DE,
                2 => Self::HL,
                3 => Self::AF,
                n => panic!("Index out of bounds: {n} > 3"),
            }
        }
    }
    impl From<u8> for R16Indirect {
        fn from(value: u8) -> Self {
            match value {
                0 => Self::BC,
                1 => Self::DE,
                2 => Self::HLI,
                3 => Self::HLD,
                n => panic!("Index out of bounds: {n} > 3"),
            }
        }
    }
    impl From<u8> for Cond {
        fn from(value: u8) -> Self {
            match value {
                0 => Self::NZ,
                1 => Self::Z,
                2 => Self::NC,
                3 => Self::C,
                n => panic!("Index out of bounds: {n} > 3"),
            }
        }
    }
}
mod disp {
    use super::*;
    use std::fmt::Display;
    impl Display for Insn {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Insn::Nop => write!(f, "nop"),
                Insn::Stop => write!(f, "stop"),
                Insn::Halt => write!(f, "halt"),

                Insn::Ld(dst, src) => write!(f, "ld\t{dst}, {src}"),
                Insn::LdImm(r8) => write!(f, "ld\t{r8}, #u8"),
                Insn::LdImm16(r16) => write!(f, "ld\t{r16}, #u16"),
                Insn::LdInd(src) => write!(f, "ld\ta, {src}"),
                Insn::StInd(dst) => write!(f, "ld\t{dst}, a"),
                Insn::LdC => write!(f, "ld\ta, [c]"),
                Insn::StC => write!(f, "ld\t[c], a"),
                Insn::LdAbs => write!(f, "ld\ta, [#u16]"),
                Insn::StAbs => write!(f, "ld\t[#u16], a"),
                Insn::StSpAbs => write!(f, "ld\t[#u16], sp"),
                Insn::LdSpHl => write!(f, "ld\tsp, hl"),
                Insn::LdHlSpRel => write!(f, "ldhl\tsp, #u8"),

                Insn::Ldh => write!(f, "ldh\ta, [0xff00 + #u8]"),
                Insn::Sth => write!(f, "ldh\t[0xff00 + #u8], a"),

                Insn::Inc(dst) => write!(f, "inc\t{dst}"),
                Insn::Inc16(dst) => write!(f, "inc\t{dst}"),
                Insn::Dec(dst) => write!(f, "dec\t{dst}"),
                Insn::Dec16(dst) => write!(f, "dec\t{dst}"),

                Insn::Rlnca => write!(f, "rlca"),
                Insn::Rrnca => write!(f, "rrca"),
                Insn::Rla => write!(f, "rla"),
                Insn::Rra => write!(f, "rra"),
                Insn::Daa => write!(f, "daa"),
                Insn::Cpl => write!(f, "cpl"),
                Insn::Scf => write!(f, "scf"),
                Insn::Ccf => write!(f, "ccf"),

                Insn::Add(src) => write!(f, "add\ta, {src}"),
                Insn::AddI => write!(f, "add\ta, #u8"),
                Insn::Add16HL(src) => write!(f, "add\tHL, {src}"),
                Insn::Add16SpI => write!(f, "add\tSP, #i8"),
                Insn::Adc(src) => write!(f, "adc\ta, {src}"),
                Insn::AdcI => write!(f, "adc\ta, #u8"),
                Insn::Sub(src) => write!(f, "sub\ta, {src}"),
                Insn::SubI => write!(f, "sub\ta, #u8"),
                Insn::Sbc(src) => write!(f, "sbc\ta, {src}"),
                Insn::SbcI => write!(f, "sbc\ta, #u8"),
                Insn::And(src) => write!(f, "and\ta, {src}"),
                Insn::AndI => write!(f, "and\ta, #u8"),
                Insn::Xor(src) => write!(f, "xor\ta, {src}"),
                Insn::XorI => write!(f, "xor\ta, #u8"),
                Insn::Or(src) => write!(f, "or\ta, {src}"),
                Insn::OrI => write!(f, "or\ta, #u8"),
                Insn::Cp(src) => write!(f, "cp\ta, {src}"),
                Insn::CpI => write!(f, "cp\ta, #u8"),

                Insn::Push(src) => write!(f, "push\t{src}"),
                Insn::Pop(dst) => write!(f, "pop\t{dst}"),

                Insn::Jr => write!(f, "jr\t#i8"),
                Insn::Jrc(cond) => write!(f, "jr\t{cond}, #i8"),
                Insn::Jp => write!(f, "jp\t#u16"),
                Insn::Jpc(cond) => write!(f, "jp\t{cond}, #u16"),
                Insn::JpHL => write!(f, "jp\tHL"),
                Insn::Call => write!(f, "call\t#u16"),
                Insn::Callc(cond) => write!(f, "call\t{cond}, #u16"),
                Insn::Di => write!(f, "di"),
                Insn::Ei => write!(f, "ei"),
                Insn::Ret => write!(f, "ret"),
                Insn::Retc(cond) => write!(f, "ret\t{cond}"),
                Insn::Reti => write!(f, "reti"),
                Insn::Rst(vector) => write!(f, "rst\t${:02x}", vector * 8),
                Insn::PrefixCB => write!(f, "; 0xCB"),
                Insn::Invalid(op) => write!(f, "_{op:02x}"),
            }
        }
    }
    impl Display for Prefixed {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Prefixed::Rlc(r8) => write!(f, "rlc\t{r8}"),
                Prefixed::Rrc(r8) => write!(f, "rrc\t{r8}"),
                Prefixed::Rl(r8) => write!(f, "rl\t{r8}"),
                Prefixed::Rr(r8) => write!(f, "rr\t{r8}"),
                Prefixed::Sla(r8) => write!(f, "sla\t{r8}"),
                Prefixed::Sra(r8) => write!(f, "sra\t{r8}"),
                Prefixed::Swap(r8) => write!(f, "swap\t{r8}"),
                Prefixed::Srl(r8) => write!(f, "srl\t{r8}"),
                Prefixed::Bit(r8, u3) => write!(f, "bit\t{r8}, {:x}", *u3 as u8),
                Prefixed::Res(r8, u3) => write!(f, "res\t{r8}, {:x}", *u3 as u8),
                Prefixed::Set(r8, u3) => write!(f, "set\t{r8}, {:x}", *u3 as u8),
            }
        }
    }
    impl Display for R8 {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                R8::B => write!(f, "b"),
                R8::C => write!(f, "c"),
                R8::D => write!(f, "d"),
                R8::E => write!(f, "e"),
                R8::H => write!(f, "h"),
                R8::L => write!(f, "l"),
                R8::HLmem => write!(f, "[hl]"),
                R8::A => write!(f, "a"),
            }
        }
    }
    impl Display for R16 {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                R16::BC => write!(f, "bc"),
                R16::DE => write!(f, "de"),
                R16::HL => write!(f, "hl"),
                R16::SP => write!(f, "sp"),
            }
        }
    }
    impl Display for R16Stack {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                R16Stack::BC => write!(f, "bc"),
                R16Stack::DE => write!(f, "de"),
                R16Stack::HL => write!(f, "hl"),
                R16Stack::AF => write!(f, "af"),
            }
        }
    }
    impl Display for R16Indirect {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                R16Indirect::BC => write!(f, "[bc]"),
                R16Indirect::DE => write!(f, "[de]"),
                R16Indirect::HLI => write!(f, "[hl+]"),
                R16Indirect::HLD => write!(f, "[hl-]"),
            }
        }
    }
    impl Display for Cond {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            match self {
                Cond::NZ => write!(f, "nz"),
                Cond::Z => write!(f, "z"),
                Cond::NC => write!(f, "nc"),
                Cond::C => write!(f, "c"),
            }
        }
    }
}
mod info {
    use super::Insn;

    impl Insn {
        /// Gets the length of the instruction in bytes
        pub fn len(&self) -> usize {
            match self {
                Insn::Nop | Insn::Stop | Insn::Halt => 1,
                Insn::Ld(_, _) => 1,
                Insn::LdImm(_) => 2,
                Insn::LdImm16(_) => 3,
                Insn::LdInd(_) | Insn::StInd(_) => 1,
                Insn::LdC | Insn::StC => 1,
                Insn::LdAbs | Insn::StAbs | Insn::StSpAbs => 3,
                Insn::LdSpHl => 1,
                Insn::LdHlSpRel => 2,
                Insn::Ldh | Insn::Sth => 2,
                Insn::Inc(_) | Insn::Inc16(_) => 1,
                Insn::Dec(_) | Insn::Dec16(_) => 1,
                Insn::Rlnca | Insn::Rrnca | Insn::Rla | Insn::Rra => 1,
                Insn::Daa | Insn::Cpl | Insn::Scf | Insn::Ccf => 1,
                Insn::Add(_) | Insn::Add16HL(_) => 1,
                Insn::AddI | Insn::Add16SpI => 2,
                Insn::Adc(_)
                | Insn::Sub(_)
                | Insn::Sbc(_)
                | Insn::And(_)
                | Insn::Xor(_)
                | Insn::Or(_)
                | Insn::Cp(_) => 1,
                Insn::AdcI
                | Insn::SubI
                | Insn::SbcI
                | Insn::AndI
                | Insn::XorI
                | Insn::OrI
                | Insn::CpI => 2,
                Insn::Push(_) | Insn::Pop(_) => 1,
                Insn::Jr | Insn::Jrc(_) => 2,
                Insn::Jp | Insn::Jpc(_) => 3,
                Insn::JpHL => 1,
                Insn::Call | Insn::Callc(_) => 3,
                Insn::Ret | Insn::Retc(_) => 1,
                Insn::Di | Insn::Ei => 1,
                Insn::Reti => 1,
                Insn::Rst(_) => 1,
                Insn::PrefixCB => 1,
                Insn::Invalid(_) => 1,
            }
        }
        #[must_use]
        pub fn is_empty(&self) -> bool {
            matches!(self, Insn::Invalid(_))
        }
    }
}
