//! Represents two 8-bit registers which may be conjoined into one 16-bit register
//!
//! The [SplitRegister] is represented by a native-endian [`Wrapping<u16>`]
//! or two [`Wrapping<u8>`]s, respectively
//!
//! # Examples
//! The [SplitRegister] can be initialized
//! ```rust
//! # use boy::cpu::split_register::SplitRegister;
//! use std::num::Wrapping;
//!
//! let reg = SplitRegister::from(0x1000);
//! assert_eq!(*reg.lo(), Wrapping(0x00));
//! assert_eq!(*reg.hi(), Wrapping(0x10));
//! ```

use std::{fmt::Debug, hash::Hash, num::Wrapping, ops::*};

use super::Flags;

/// The low-byte index constant
const LO: usize = if 1u16.to_be() != 1u16 { 0 } else { 1 };
/// The high-byte index constant
const HI: usize = 1 - LO;

#[derive(Clone, Copy)]
#[repr(C)]
pub union SplitRegister {
    lh: [Wrapping<u8>; 2],
    af: [Flags; 2],
    r16: Wrapping<u16>,
}
impl SplitRegister {
    /// Read the low byte of the register
    pub fn lo(&self) -> &Wrapping<u8> {
        // SAFETY: The GB's registers store POD types, so it's safe to transmute
        unsafe { &self.lh[LO] }
    }
    /// Read the high byte of the register
    pub fn hi(&self) -> &Wrapping<u8> {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &self.lh[HI] }
    }
    /// Read the [flags byte](Flags) of the register
    pub fn flags(&self) -> &Flags {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &self.af[LO] }
    }
    /// Read the contents of the combined register
    pub fn wide(&self) -> &Wrapping<u16> {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &self.r16 }
    }

    /// Get a mutable reference to the low byte of the register
    pub fn lo_mut(&mut self) -> &mut Wrapping<u8> {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &mut self.lh[LO] }
    }
    /// Gets a mutable reference to the high byte of the register
    pub fn hi_mut(&mut self) -> &mut Wrapping<u8> {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &mut self.lh[HI] }
    }
    /// Gets a mutable reference to the [flags byte](Flags) of the register
    pub fn flags_mut(&mut self) -> &mut Flags {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &mut self.af[LO] }
    }
    /// Gets a mutable reference to the combined register
    pub fn wide_mut(&mut self) -> &mut Wrapping<u16> {
        // SAFETY: See [SplitRegister::lo]
        unsafe { &mut self.r16 }
    }

    /// Constructs a SplitRegister from an array of big-endian bytes.
    pub fn from_be_bytes(value: [u8; 2]) -> Self {
        Self {
            lh: [Wrapping(value[HI]), Wrapping(value[LO])],
        }
    }
    /// Constructs a SplitRegister from an array of little-endian bytes.
    pub fn from_le_bytes(value: [u8; 2]) -> Self {
        Self {
            lh: [Wrapping(value[LO]), Wrapping(value[HI])],
        }
    }
    /// Constructs a SplitRegister from an array of host-endian bytes.
    pub fn from_ne_bytes(value: [u8; 2]) -> Self {
        Self {
            lh: [Wrapping(value[0]), Wrapping(value[1])],
        }
    }
}
impl Debug for SplitRegister {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "[{:02x}, {:02x}]", self.lo(), self.hi())
    }
}
impl Default for SplitRegister {
    fn default() -> Self {
        Self { r16: Wrapping(0) }
    }
}
impl PartialEq for SplitRegister {
    fn eq(&self, other: &Self) -> bool {
        self.wide().eq(other.wide())
    }
}
impl Eq for SplitRegister {}
impl PartialOrd for SplitRegister {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}
impl Ord for SplitRegister {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.wide().cmp(other.wide())
    }
}
impl Hash for SplitRegister {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.wide().hash(state)
    }
}
impl From<u16> for SplitRegister {
    fn from(value: u16) -> Self {
        Self {
            r16: Wrapping(value),
        }
    }
}
impl From<Wrapping<u16>> for SplitRegister {
    fn from(value: Wrapping<u16>) -> Self {
        Self { r16: value }
    }
}

macro impl_ops(for $base:ident {$($trait:ident<$($ty:ty),*$(,)?> $func:ident()),*$(,)?}) {
    $(
        impl $trait<Wrapping<u16>> for $base {
            type Output = Self;
            fn $func(self, rhs: Wrapping<u16>) -> Self::Output {
                Self { r16: self.wide().$func(rhs) }
            }
        }
        impl $trait<Self> for $base {
            type Output = Self;
            fn $func(self, rhs: Self) -> Self::Output {
                self.$func(*rhs.wide())
            }
        }
        $(impl $trait<$ty> for $base {
            type Output = Self;
            fn $func(self, rhs: $ty) -> Self::Output {
                self.$func(Wrapping(rhs as u16))
            }
        })*
    )*
}
impl_ops!(for SplitRegister {
    Mul<u8, u16> mul(),
    Div<u8, u16> div(),
    Rem<u8, u16> rem(),
    Add<u8, u16> add(),
    Sub<u8, u16> sub(),
    BitAnd<u8, u16> bitand(),
    BitOr<u8, u16>  bitor(),
    BitXor<u8, u16> bitxor(),
});

#[test]
fn low_is_low() {
    let reg = SplitRegister {
        r16: Wrapping(0x00ff),
    };
    assert_eq!(*reg.lo(), Wrapping(0xff))
}
#[test]
fn hi_is_hi() {
    let reg = SplitRegister {
        r16: Wrapping(0xff00),
    };
    assert_eq!(*reg.hi(), Wrapping(0xff))
}
#[test]
fn from_bytes() {
    let be = SplitRegister::from_be_bytes([0xc5, 0x77]);
    let le = SplitRegister::from_le_bytes([0xc5, 0x77]);
    let ne = SplitRegister::from_ne_bytes(0x1234u16.to_ne_bytes());
    assert_eq!(*be.wide(), Wrapping(0xc577));
    assert_eq!(*le.wide(), Wrapping(0x77c5));
    assert_eq!(*ne.wide(), Wrapping(0x1234));
}
