//! RAM, CROM, and cartridge mappers
//!
//! # Gameboy Color memory map:
//! - `0000..4000` => ROM bank 0
//! - `4000..8000` => ROM banks 1-N (via [Mapper])
//! - `8000..A000` => VRAM Bank 0-1
//! - `A000..C000` => Cart RAM (via [Mapper])
//! - `C000..D000` => WRAM Bank 0
//! - `D000..E000` => WRAM Bank 1-7
//! - `E000..FE00` => ECHO RAM (WRAM Bank 0)
//! - `FE00..FEA0` => Object Attribute Memory (OAM)
//! - `FEA0..FF00` => Prohibited region (corrupts OAM on DMG)
//! - `FF00..FF80` => [Memory mapped I/O][MMIO]
//! - `FF80..FFFF` => HiRAM
//! - `FFFF` => Interrupt Enable Register
//!
//! [MMIO]: https://gbdev.io/pandocs/Hardware_Reg_List.html

use self::{
    banked::{Banked, UpperBanked},
    io::BusIO,
    mapper::Mapper,
};
pub mod banked;
pub mod io;

/// The bus operating mode
#[derive(Clone, Copy, Debug)]
pub enum Mode {
    DMG,
    CGB,
}
impl Mode {
    pub fn from_bits(value: u8) -> Self {
        match value & 0x40 {
            0 => Self::DMG,
            _ => Self::CGB,
        }
    }
    pub fn to_bits(self) -> u8 {
        (self as u8) << 7
    }
}

#[derive(Debug)]
pub struct Bus {
    mode: Mode,
    // TODO: define a way to have an arbitrary memory map
    cart: Cart,
    // VRAM is a 0x2000 B window from 0x8000..0xa000, with two banks
    vram: Banked<0x8000, 0x2000, 2>,
    // WRAM is a 0x2000 B window from 0xc000..0xe000. The upper half is banked.
    wram: UpperBanked<0xc000, 0x1000, 8>,
    // HRAM is a 0x80 B window from 0xff80..0xffff
    hram: [u8; 0x80],
    // Joypad driver
    // Serial driver
    // Timer and divider
    // Interrupt controller
    // Audio controller
    // PPU/LCD controller
    mmio: [u8; 0x80],
    //
}

impl BusIO for Bus {
    fn read(&self, addr: usize) -> Option<u8> {
        match addr {
            // Cart ROM
            0x0000..=0x7fff => self.cart.read(addr),
            // VRAM
            0x8000..=0x9fff => self.vram.read(addr),
            // Cart RAM
            0xa000..=0xbfff => self.cart.read(addr),
            // Work RAM
            0xc000..=0xdfff => self.wram.read(addr),
            // Wram mirror
            0xe000..=0xfdff => self.wram.read(addr - 0x2000),
            // Graphics OAM
            0xfe00..=0xfe9f => Some(0x0a),
            // Illegal
            0xfea0..=0xfeff => None,
            // Memory mapped IO
            0xff00..=0xff7f => self.mmio.read(addr & 0x7f),
            // HiRAM
            0xff80..=0xffff => self.hram.read(addr & 0x7f),
            _ => None,
        }
    }

    fn write(&mut self, addr: usize, data: u8) -> Option<()> {
        match addr {
            // Cart ROM
            0x0000..=0x7fff => self.cart.write(addr, data),
            // VRAM
            0x8000..=0x9fff => self.vram.write(addr, data),
            // Cart RAM
            0xa000..=0xbfff => self.cart.write(addr, data),
            // Work RAM
            0xc000..=0xdfff => self.wram.write(addr, data),
            // Wram mirror
            0xe000..=0xfdff => self.wram.write(addr - 0x2000, data),
            // Graphics OAM
            0xfe00..=0xfe9f => Some(()),
            // Illegal
            0xfea0..=0xfeff => None,
            // Memory mapped IO
            0xff00..=0xff7f => self.mmio.write(addr & 0x7f, data),
            // HiRAM
            0xff80..=0xffff => self.hram.write(addr & 0x7f, data),
            _ => None,
        }
    }
}

impl Bus {
    pub fn new(cart: Cart) -> Self {
        Self {
            cart,
            mode: Mode::CGB,
            vram: Default::default(),
            wram: Default::default(),
            mmio: [0; 128],
            hram: [0; 128],
        }
    }
    pub fn mmio_read(&self, addr: usize) -> Option<u8> {
        match addr {
            // Joypad I/O
            0xff00 => eprintln!("DMG Joypad output"),
            // DMG/CGB Serial transfer
            0xff01..=0xff02 => eprintln!("DMG Serial transfer"),
            // DMG Timer and divider
            0xff04..=0xff07 => eprintln!("DMG Timer and divider"),
            // Interrupt controller (IF)
            0xff0f => eprintln!("Interrupt controller"),
            // DMG Audio
            0xff10..=0xff26 => eprintln!("DMG Audio"),
            // DMG Wave pattern memory
            0xff30..=0xff3f => eprintln!("DMG Wave pattern memory"),
            // DMG LCD control, status, position, scrolling, and palettes
            0xff40..=0xff4b => {
                eprintln!("DMG LCD control, status, position, scrolling, and palettes")
            }
            // Disable bootrom
            0xff50 => eprintln!("Set to non-zero to disable bootrom"),
            _ => match self.mode {
                Mode::CGB => {
                    if let Some(data) = self.mmio_read_cgb(addr) {
                        return Some(data);
                    }
                }
                Mode::DMG => {}
            },
        }
        self.mmio.get(addr % 0x80).copied()
    }

    pub fn mmio_read_cgb(&self, addr: usize) -> Option<u8> {
        match addr {
            // CGB Key1 speed switch
            0xff4d => eprintln!("CGB Key1"),
            // CGB VRam bank select
            0xff4f => return Some(self.vram.selected() as u8 | 0xfc),
            // CGB Vram DMA controller
            0xff51..=0xff55 => eprintln!("CGB Vram DMA"),
            // CGB Infrared
            0xff56 => eprintln!("Infrared port"),
            // CGB BG/OBJ palettes
            0xff68..=0xff6b => eprintln!("CGB BG/Obj palettes"),
            // CGB WRam bank select
            0xff70 => return Some(self.wram.selected() as u8 | 0xfe),
            // CGB Undocumented registers
            0xff72..=0xff75 => return Some(00),
            // CGB PCM Read registers
            0xff76..=0xff77 => eprintln!("CGB PCM Read Registers"),
            _ => {}
        }
        None
    }

    pub fn mmio_write(&mut self, addr: usize, data: u8) -> Option<()> {
        match addr {
            // Joypad I/O
            0xff00 => eprintln!("DMG Joypad output"),
            // DMG/CGB Serial transfer
            0xff01..=0xff02 => eprintln!("DMG Serial transfer"),
            // DMG Timer and divider
            0xff04..=0xff07 => eprintln!("DMG Timer and divider"),
            // Interrupt controller (IF)
            0xff0f => eprintln!("Interrupt controller"),
            // DMG Audio
            0xff10..=0xff26 => eprintln!("DMG Audio"),
            // DMG Wave pattern memory
            0xff30..=0xff3f => eprintln!("DMG Wave pattern memory"),
            // DMG LCD control, status, position, scrolling, and palettes
            0xff40..=0xff4b => {
                eprintln!("DMG LCD control, status, position, scrolling, and palettes")
            }
            // Disable bootrom
            0xff50 => todo!("Set to non-zero to disable bootrom"),
            _ => match self.mode {
                Mode::CGB => {
                    self.mmio_write_cgb(addr, data);
                }
                Mode::DMG => {}
            },
        }
        self.mmio.write(addr % 0x80, data)
    }

    pub fn mmio_write_cgb(&mut self, addr: usize, data: u8) -> Option<()> {
        match addr {
            // CGB Key1 speed switch
            0xff4d => eprintln!("CGB Key1"),
            // CGB VRam bank select
            0xff4f => {
                self.vram.select(data as usize);
                return Some(());
            }
            // CGB Vram DMA controller
            0xff51..=0xff55 => eprintln!("CGB Vram DMA"),
            // CGB Infrared
            0xff56 => eprintln!("Infrared port"),
            // CGB BG/OBJ palettes
            0xff68..=0xff6b => eprintln!("CGB BG/Obj palettes"),
            // CGB WRam bank select
            0xff70 => {
                self.wram.select(data as usize);
                return Some(());
            }
            // CGB Undocumented registers
            0xff72..=0xff75 => return Some(()),
            // CGB PCM Read registers
            0xff76..=0xff77 => eprintln!("CGB PCM Read Registers"),
            _ => {}
        }
        None
    }
}

pub struct Cart {
    pub rom: Vec<u8>,
    pub ram: Vec<u8>,
    mapper: Box<dyn Mapper>,
}

mod cart {
    use crate::memory::io::BusAux;

    use super::{mapper::*, BusIO, Cart};
    use std::fmt::Debug;
    impl BusIO for Cart {
        fn read(&self, addr: usize) -> Option<u8> {
            match self.mapper.read(addr) {
                Response::None => None,
                Response::Data(data) => Some(data),
                Response::Ram(addr) => self.ram.get(addr).copied(),
                Response::Rom(addr) => self.rom.get(addr).copied(),
            }
        }

        fn write(&mut self, addr: usize, data: u8) -> Option<()> {
            match self.mapper.write(addr, data) {
                Response::None => None,
                Response::Data(_) => Some(()),
                Response::Ram(addr) => self.ram.get_mut(addr).map(|rambyte| *rambyte = data),
                Response::Rom(addr) => {
                    panic!("Attempted to write to ROM address :{addr:x}")
                }
            }
        }
    }
    /// Read cartridge header, as defined in the [Pan docs](https://gbdev.io/pandocs/The_Cartridge_Header.html)
    impl Cart {
        pub fn new(rom: Vec<u8>) -> Self {
            let rom_size = rom.romsize().expect("ROM should have cart header!");
            let ram_size = rom.ramsize().expect("ROM should have cart header!");
            let cart_type = rom.carttype().expect("ROM should have cart header!");
            Self {
                mapper: match cart_type {
                    0 => Box::<no_mbc::NoMBC>::default(),
                    1 => Box::new(mbc1::MBC1::new(rom_size, 0)), // ROM
                    2 => Box::new(mbc1::MBC1::new(rom_size, ram_size)), // ROM + RAM
                    3 => Box::new(mbc1::MBC1::new(rom_size, ram_size)), // ROM + RAM + Battery
                    // _ => Box::<NoMBC>::default(),
                    n => todo!("Mapper 0x{n:02x}"),
                },
                rom,
                ram: vec![],
            }
        }
    }

    impl Debug for Cart {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            f.debug_struct("Cart")
                .field("rom", &self.rom)
                .field("ram", &self.ram)
                .finish_non_exhaustive()
        }
    }

    impl From<Vec<u8>> for Cart {
        fn from(value: Vec<u8>) -> Self {
            Self::new(value)
        }
    }
    impl From<&[u8]> for Cart {
        fn from(value: &[u8]) -> Self {
            Self::new(Vec::from(value))
        }
    }

    impl<Item: AsRef<u8>> FromIterator<Item> for Cart {
        fn from_iter<T: IntoIterator<Item = Item>>(iter: T) -> Self {
            Self::new(iter.into_iter().map(|item| *(item.as_ref())).collect())
        }
    }

    impl std::fmt::Display for Cart {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            write!(f, "Entry:     {:02x?}", self.entry())?;
            // if let Some(logo) = self.logo() {
            //     write!(f, "Logo:      ")?;
            //     for byte in logo {
            //         write!(f, "{byte:02x}")?;
            //     }
            //     writeln!(f)?;
            // }
            write!(f, "Title:     {:02x?}", self.title())?;
            write!(f, "MFCode:    {:02x?}", self.mfcode())?;
            write!(f, "CGB Flag:  {:02x?}", self.cgbflag())?;
            write!(f, "New Lic:   {:02x?}", self.newlicensee())?;
            write!(f, "SGB Flag:  {:02x?}", self.sgbflag())?;
            write!(f, "Cart Type: {:02x?}", self.carttype())?;
            write!(f, "ROM Size:  {:02x?}", self.romsize())?;
            write!(f, "RAM Size:  {:02x?}", self.ramsize())?;
            write!(f, "Dest Code: {:02x?}", self.destcode())?;
            write!(f, "Old Lic:   {:02x?}", self.oldlicencee())?;
            write!(f, "ROM Ver:   {:02x?}", self.maskromver())?;
            write!(f, "H Cksum:   {:02x?}", self.headercksum())?;
            write!(f, "G Cksum:   {:04x?}", self.globalcksum())
        }
    }
}

pub mod mapper {
    #[derive(Clone, Copy, Debug, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
    pub enum Response {
        /// No data to return
        #[default]
        None,
        /// [Mapper] returned some data directly
        Data(u8),
        /// [Mapper] mapped incoming address to cart RAM
        Ram(usize),
        /// [Mapper] mapped incoming address to cart ROM
        Rom(usize),
    }

    pub trait Mapper {
        fn read(&self, addr: usize) -> Response;
        fn write(&mut self, addr: usize, val: u8) -> Response;
    }

    pub mod no_mbc {
        use super::*;
        #[derive(Clone, Copy, Debug, Default)]
        pub struct NoMBC;
        impl Mapper for NoMBC {
            fn read(&self, addr: usize) -> Response {
                match addr {
                    0x0000..=0x7fff => Response::Rom(addr),
                    0xa000..=0xbfff => Response::Ram(addr - 0x8000),
                    _ => Response::None,
                }
            }

            fn write(&mut self, addr: usize, _val: u8) -> Response {
                match addr {
                    0xa000..=0xbfff => Response::Ram(addr - 0xA000),
                    _ => Response::None,
                }
            }
        }
    }

    pub mod mbc1 {
        use super::*;

        #[derive(Clone, Copy, Debug, Default)]
        pub struct MBC1 {
            rom_mask: u8,
            ram_mask: u8,
            mode: u8,
            rom_bank: u8,
            ram_bank: u8,
            ram_enable: bool,
        }

        impl MBC1 {
            pub fn new(rom_size: usize, ram_size: usize) -> Self {
                Self {
                    rom_mask: (rom_size).wrapping_sub(1) as u8,
                    ram_mask: (ram_size).wrapping_sub(1) as u8,
                    rom_bank: 1,
                    ..Default::default()
                }
            }
            pub fn rom_bank_lower(&self) -> usize {
                (((self.ram_bank << 5) & self.rom_mask) as usize) << 14
            }
            pub fn rom_bank_upper(&self) -> usize {
                ((self.rom_bank | (self.ram_bank << 5) & self.rom_mask) as usize) << 14
            }
            pub fn ram_bank_base(&self) -> usize {
                ((self.ram_bank & self.ram_mask) as usize) << 13
            }
        }

        impl Mapper for MBC1 {
            fn read(&self, addr: usize) -> Response {
                match (addr, self.mode) {
                    (0x0000..=0x3fff, 0) => Response::Rom(addr),
                    (0x0000..=0x3fff, _) => Response::Rom(self.rom_bank_lower() | (addr & 0x3fff)),

                    (0x4000..=0x7fff, _) => Response::Rom(self.rom_bank_upper() | (addr & 0x3fff)),

                    (0xa000..=0xbfff, 0) => Response::Ram(addr & 0x1fff),
                    (0xa000..=0xbfff, _) => Response::Ram(self.ram_bank_base() | (addr & 0x1fff)),

                    _ => Response::None,
                }
            }

            fn write(&mut self, addr: usize, val: u8) -> Response {
                match (addr, self.mode) {
                    (0x0000..=0x1fff, _) => {
                        self.ram_enable = val & 0xf == 0xa;
                    }
                    (0x2000..=0x3fff, _) => {
                        self.rom_bank = (val & 0x1f).max(1);
                    }
                    (0x4000..=0x5fff, _) => {
                        self.ram_bank = val & 0x3;
                    }
                    (0x6000..=0x7fff, _) => {
                        self.mode = val & 1;
                    }
                    (0xa000..=0xbfff, 0) if self.ram_enable => {
                        return Response::Ram(addr & 0x1fff);
                    }
                    (0xa000..=0xbfff, 1) if self.ram_enable => {
                        return Response::Ram((self.ram_bank as usize & 0b11 << 13) | addr & 0x1fff)
                    }
                    _ => return Response::None,
                }
                Response::Data(0)
            }
        }
    }
}
