//! Shares the upper half of a memory region among a number of memory banks
use super::io::BusIO;

/// Shares an entire memory region from BASE to BASE + LEN among multiple banks
#[derive(Clone, Debug)]
pub struct Banked<const BASE: usize, const LEN: usize, const BANKS: usize> {
    selected: usize,
    mem: [[u8; LEN]; BANKS],
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> Banked<BASE, LEN, BANKS> {
    /// Creates a new [Banked<BASE, LEN, BANKS>](Banked)
    pub fn new() -> Self {
        Self::default()
    }
    /// Selects a bank between 0 and `LEN` - 1, saturating
    pub fn select(&mut self, bank: usize) {
        self.selected = bank.min(BANKS - 1)
    }
    pub fn selected(&self) -> usize {
        self.selected
    }
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> Default for Banked<BASE, LEN, BANKS> {
    #[rustfmt::skip]
    fn default() -> Self {
        Self { selected: 0, mem: [[0; LEN]; BANKS] }
    }
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> BusIO for Banked<BASE, LEN, BANKS> {
    fn read(&self, addr: usize) -> Option<u8> {
        let addr = addr.checked_sub(BASE).filter(|v| *v < LEN)?;
        self.mem[self.selected].read(addr)
    }

    fn write(&mut self, addr: usize, data: u8) -> Option<()> {
        let addr = addr.checked_sub(BASE).filter(|v| *v < LEN)?;
        self.mem[self.selected].write(addr, data)
    }
}

impl<const A: usize, const LEN: usize, const BANKS: usize> AsRef<[u8]> for Banked<A, LEN, BANKS> {
    fn as_ref(&self) -> &[u8] {
        //! Adapted from [core::slice::flatten](https://doc.rust-lang.org/std/primitive.slice.html#method.flatten)
        let slice = self.mem.as_slice().as_ptr().cast();
        // SAFETY:
        // - `LEN * BANKS` cannot overflow because `self` is already in the address space.
        // - `[T]` is layout-identical to `[T; N]`
        unsafe { core::slice::from_raw_parts(slice, LEN * BANKS) }
    }
}

impl<const A: usize, const LEN: usize, const BANKS: usize> AsMut<[u8]> for Banked<A, LEN, BANKS> {
    fn as_mut(&mut self) -> &mut [u8] {
        let slice = self.mem.as_mut_slice().as_mut_ptr().cast();
        // SAFETY: see [Banked::as_ref]
        unsafe { core::slice::from_raw_parts_mut(slice, LEN * BANKS) }
    }
}

// --------------------------------- //

/// Shares the region from `BASE` + `LEN` to `BASE` + 2 * `LEN` among `BANKS`-1 banks,
/// while keeping the lower bank constant
/// # Constants:
/// - `BASE`: The location where the [UpperBanked] begins
/// - `LEN`: The length of one bank. This is *half* of the in-memory length,
/// - `BANKS`: The number of banks
#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct UpperBanked<const BASE: usize, const LEN: usize, const BANKS: usize> {
    selected: usize,
    mem: [[u8; LEN]; BANKS],
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> UpperBanked<BASE, LEN, BANKS> {
    /// Creates a new [`UpperBanked<BASE, LEN, BANKS>`](Banked)
    pub fn new() -> Self {
        Self::default()
    }
    /// Selects a bank between 1 and `LEN` - 1, saturating to `LEN` - 1
    pub fn select(&mut self, bank: usize) {
        self.selected = bank.clamp(1, BANKS - 1)
    }
    /// Gets the currently selected upper bank
    pub fn selected(&self) -> usize {
        self.selected
    }
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> Default
    for UpperBanked<BASE, LEN, BANKS>
{
    #[rustfmt::skip]
    fn default() -> Self {
        Self { selected: 1, mem: [[0; LEN]; BANKS] }
    }
}

impl<const BASE: usize, const LEN: usize, const BANKS: usize> BusIO
    for UpperBanked<BASE, LEN, BANKS>
{
    fn read(&self, addr: usize) -> Option<u8> {
        // Pray this gets optimized
        let addr = addr.checked_sub(BASE).filter(|v| *v < LEN * 2)?;
        let (bank, addr) = (addr / LEN, addr % LEN);
        match bank {
            0 => self.mem[0].read(addr),
            _ => self.mem[self.selected].read(addr),
        }
    }

    fn write(&mut self, addr: usize, data: u8) -> Option<()> {
        // Pray this gets optimized
        let addr = addr.checked_sub(BASE).filter(|v| *v < LEN * 2)?;
        let (bank, addr) = (addr / LEN, addr % LEN);
        match bank {
            0 => self.mem[0].write(addr, data),
            _ => self.mem[self.selected].write(addr, data),
        }
    }
}

impl<const A: usize, const LEN: usize, const BANKS: usize> AsRef<[u8]>
    for UpperBanked<A, LEN, BANKS>
{
    fn as_ref(&self) -> &[u8] {
        let slice_ptr = self.mem.as_slice().as_ptr().cast();
        // SAFETY: see [Banked::as_ref]
        unsafe { core::slice::from_raw_parts(slice_ptr, LEN * BANKS) }
    }
}

impl<const A: usize, const LEN: usize, const BANKS: usize> AsMut<[u8]>
    for UpperBanked<A, LEN, BANKS>
{
    fn as_mut(&mut self) -> &mut [u8] {
        let slice_ptr = self.mem.as_mut_slice().as_mut_ptr().cast();
        // SAFETY: see [UpperBanked::as_ref]
        unsafe { core::slice::from_raw_parts_mut(slice_ptr, LEN * BANKS) }
    }
}

#[cfg(test)]
mod test {
    use super::UpperBanked;
    #[test]
    fn upper_banked_asref_slice() {
        let mut ub = UpperBanked::<0, 10, 20>::new();
        let mut counter = 0u8;
        for bank in &mut ub.mem {
            for addr in bank {
                *addr = counter;
                counter += 1;
            }
        }
        for (idx, data) in ub.as_ref().iter().enumerate() {
            assert_eq!(idx, *data as usize)
        }
    }
    #[test]
    fn upper_banked_bounds() {
        let ub = UpperBanked::<0, 8, 32>::new();
        assert_eq!(ub.as_ref().len(), 8 * 32);
    }
}
